﻿using FireDrill.Windows;
using Microsoft.Kinect;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Media;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace FireDrill
{
    /// <summary>
    /// MainWindow.xaml 的交互逻辑
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        // 活动的 Kinect 传感器
        private KinectSensor _kinectSensor = null;

        // 身体框架索引渲染器
        private BodyIndexFrameReader bodyIndexFrameReader = null;
        // 身体框架中包含的数据的说明
        private FrameDescription bodyIndexFrameDescription = null;
        // 要显示的位图
        private WriteableBitmap bodyIndexBitmap = null;
        // 帧数据转换成颜色的中间存储器
        private uint[] bodyIndexPixels = null;
        // 位图中 RGB 像素的大小
        private const int BytesPerPixel = 4;
        // 用于显示 BodyIndexFrame 数据的颜色集合
        private static readonly uint[] BodyColor =
        {
            //0x0000FF00,
            //0x00FF0000,
            //0xFFFF4000,
            //0x40FFFF00,
            //0xFF40FF00,
            //0xFF808000,
            (uint)System.Drawing.Color.White.ToArgb(),
            (uint)System.Drawing.Color.White.ToArgb(),
            (uint)System.Drawing.Color.White.ToArgb(),
            (uint)System.Drawing.Color.White.ToArgb(),
            (uint)System.Drawing.Color.White.ToArgb(),
            (uint)System.Drawing.Color.White.ToArgb(),
    };

        // 坐标映射器将一种类型的点映射到另一种
        private CoordinateMapper _coordinateMapper = null;
        // 身体框架渲染器
        private BodyFrameReader _bodyFrameReader = null;
        // 颜色框架渲染器
        private ColorFrameReader _colorFrameReader = null;
        // 身体的数组
        private Body[] _bodies = null;
        // 骨骼的定义
        private List<Tuple<JointType, JointType>> _bones;
        // 追踪的每个身体的颜色列表
        private List<Pen> _bodyColors;
        // 用于体绘制输出的绘图组
        private DrawingGroup _drawingGroup;
        // 显示的绘图图像
        private DrawingImage _imageSource;
        // 显示宽度，高度（深度空间）
        private int _displayWidth, _displayHeight;
        // 绘制的手圆半径
        private const double HandSize = 30;
        // 绘制接合线的厚度
        private const double JointThickness = 3;
        // 夹边矩形的厚度
        private const double ClipBoundsThickness = 10;
        // 将相机空间点的Z值从负值夹持的常量
        private const float InferredZPositionClamp = 0.1f;
        // 用于绘制当前跟踪为闭合的手的画笔
        private readonly Brush _handClosedBrush = new SolidColorBrush(Color.FromArgb(128, 255, 0, 0));
        // 用于绘制当前跟踪为打开状态的手的画笔
        private readonly Brush _handOpenBrush = new SolidColorBrush(Color.FromArgb(128, 0, 255, 0));
        // 用于绘制当前在套索（指针）位置跟踪的手的画笔
        private readonly Brush _handLassoBrush = new SolidColorBrush(Color.FromArgb(128, 0, 0, 255));
        // 用于绘制当前跟踪的关节的画笔
        private readonly Brush _trackedJointBrush = new SolidColorBrush(Color.FromArgb(255, 68, 192, 68));
        // 用于绘制当前推断的关节的画笔     
        private readonly Brush _inferredJointBrush = Brushes.Yellow;
        // 用于绘制当前推断的骨骼的笔      
        private readonly Pen _inferredBonePen = new Pen(Brushes.Gray, 1);
        // 要显示的当前状态文本
        private string _statusText;
        // 要显示的提示文本
        private string _tipText;
        // 阈值（1/2框）（单位：米）
        private const float Threshold = 0.15f;
        // 当前状态
        private int _gesture = -1;
        // Bitmap to display
        private WriteableBitmap _colorBitmap;
        private SoundPlayer soundPlayer = new SoundPlayer();

        // INotifyPropertyChangedPropertyChanged 事件允许窗口控件绑定到可更改的数据
        public event PropertyChangedEventHandler PropertyChanged;

        public string StatusText
        {
            get
            {
                return _statusText;
            }

            set
            {
                if (_statusText == value) return;
                _statusText = value;
                // 通知任何绑定元素文本已更改
                PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("StatusText"));
            }
        }

        public ImageSource ImageSource
        {
            get
            {
                return bodyIndexBitmap;
            }
        }

        public MainWindow()
        {
            // 初始化窗口的组件（控件）
            InitializeComponent();
            // 目前支持一个传感器
            _kinectSensor = KinectSensor.GetDefault();
            // open the reader for the depth frames
            bodyIndexFrameReader = _kinectSensor.BodyIndexFrameSource.OpenReader();
            // wire handler for frame arrival
            bodyIndexFrameReader.FrameArrived += Reader_FrameArrived;
            bodyIndexFrameDescription = _kinectSensor.BodyIndexFrameSource.FrameDescription;
            // allocate space to put the pixels being converted
            bodyIndexPixels = new uint[bodyIndexFrameDescription.Width * bodyIndexFrameDescription.Height];
            // create the bitmap to display
            bodyIndexBitmap = new WriteableBitmap(bodyIndexFrameDescription.Width, bodyIndexFrameDescription.Height, 96.0, 96.0, PixelFormats.Bgra32, null);
            // 设置 IsAvailableChanged 事件通知程序
            _kinectSensor.IsAvailableChanged += this.Sensor_IsAvailableChanged;
            // 打开传感器
            _kinectSensor.Open();
            // set the status text
            StatusText = _kinectSensor.IsAvailable ? Properties.Resources.RunningStatusText : Properties.Resources.NoSensorStatusText;
            // 在这个简单的例子中使用window对象作为视图模型
            DataContext = this;
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            // 交互式控制
            VideoPlayer.LoadedBehavior = MediaState.Manual;
            // 添加元素加载完成事件 -- 自动开始播放
            VideoPlayer.Loaded += new RoutedEventHandler(Media_Loaded);
            // 添加媒体播放结束事件 -- 重新播放
            VideoPlayer.MediaEnded += new RoutedEventHandler(Media_MediaEnded);
            // 添加元素卸载完成事件 -- 停止播放
            VideoPlayer.Unloaded += new RoutedEventHandler(Media_Unloaded);
            Assembly assembly = Assembly.GetExecutingAssembly();
            string videoContent = Properties.Settings.Default.Video;
            string audioContent = Properties.Settings.Default.Audio;
            if (string.IsNullOrEmpty(videoContent))
            {
                MessageBox.Show("请到设置页面配置视频！", "提示", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            else
            {
                // 绑定视频文件
                VideoPlayer.Source = new Uri(videoContent);
                VideoPlayer.Play();
            }
            if (string.IsNullOrEmpty(audioContent))
            {
                MessageBox.Show("请到设置页面配置音频！", "提示", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            else
            {
                soundPlayer.SoundLocation = audioContent;
            }
        }

        private void Window_Closing(object sender, CancelEventArgs e)
        {
            var messageBoxResult = MessageBox.Show("确定退出吗？", "提示", MessageBoxButton.OKCancel, MessageBoxImage.Question);
            if (messageBoxResult == MessageBoxResult.Cancel)
            {
                e.Cancel = true;
            }
            else
            {
                soundPlayer?.Stop();
                if (bodyIndexFrameReader != null)
                {
                    // 移除事件处理程序
                    bodyIndexFrameReader.FrameArrived -= Reader_FrameArrived;
                    // 清除身体框架索引渲染器
                    bodyIndexFrameReader.Dispose();
                    bodyIndexFrameReader = null;
                }
                if (_bodyFrameReader != null)
                {
                    _bodyFrameReader.Dispose();
                    _bodyFrameReader = null;
                }

                if (_colorFrameReader != null)
                {
                    _colorFrameReader.Dispose();
                    _colorFrameReader = null;
                }

                if (_kinectSensor != null)
                {
                    _kinectSensor.Close();
                    _kinectSensor = null;
                }
                Application.Current.Shutdown();
            }
        }

        private void Window_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.Key)
            {
                case Key.Escape:
                    ToolBar.Visibility = Visibility.Visible;
                    VideoPlayer.Margin = new Thickness { Left = 0, Top = 32, Right = 0, Bottom = 0 };
                    ViewBox.Height = 150;
                    ViewBox.Width = 150;
                    WindowState = WindowState.Normal;
                    WindowStyle = WindowStyle.SingleBorderWindow;
                    break;
                case Key.Space:
                    //var bitmapImage = new BitmapImage(new Uri("I:\\图片\\background.jpg", UriKind.Absolute));
                    //_colorBitmap = new WriteableBitmap(bitmapImage);
                    //StatusText = DateTime.Now.ToLongTimeString();
                    break;
            }
        }

        private void Reader_FrameArrived(object sender, BodyIndexFrameArrivedEventArgs e)
        {
            bool bodyIndexFrameProcessed = false;
            using (BodyIndexFrame bodyIndexFrame = e.FrameReference.AcquireFrame())
            {
                if (bodyIndexFrame != null)
                {
                    // 处理 body 索引数据的最快方法是直接访问底层缓冲区
                    using (KinectBuffer bodyIndexBuffer = bodyIndexFrame.LockImageBuffer())
                    {
                        // 验证数据并将颜色数据写入显示位图
                        if (((bodyIndexFrameDescription.Width * bodyIndexFrameDescription.Height) == bodyIndexBuffer.Size) &&
                            (bodyIndexFrameDescription.Width == bodyIndexBitmap.PixelWidth) && (bodyIndexFrameDescription.Height == bodyIndexBitmap.PixelHeight))
                        {
                            ProcessBodyIndexFrameData(bodyIndexBuffer.UnderlyingBuffer, bodyIndexBuffer.Size);
                            bodyIndexFrameProcessed = true;
                        }
                    }
                }
            }

            if (bodyIndexFrameProcessed)
            {
                // RenderBodyIndexPixels
                bodyIndexBitmap.WritePixels(
                new Int32Rect(0, 0, bodyIndexBitmap.PixelWidth, bodyIndexBitmap.PixelHeight),
                bodyIndexPixels,
                bodyIndexBitmap.PixelWidth * BytesPerPixel,
                0);
            }
        }

        private unsafe void ProcessBodyIndexFrameData(IntPtr bodyIndexFrameData, uint bodyIndexFrameDataSize)
        {
            byte* frameData = (byte*)bodyIndexFrameData;
            // 将正文索引转换为可视表示
            for (int i = 0; i < (int)bodyIndexFrameDataSize; ++i)
            {
                // BodyColor 数组的大小已调整为与 BodyFrameSource.BodyCount 匹配
                if (frameData[i] < BodyColor.Length)
                {
                    // 像素是播放器的一部分，显示适当的颜色
                    bodyIndexPixels[i] = BodyColor[frameData[i]];
                }
                else
                {
                    // 此像素不是播放器的一部分，显示黑色
                    //bodyIndexPixels[i] = 0x0000000F;
                    bodyIndexPixels[i] = (uint)System.Drawing.Color.Transparent.ToArgb();
                }
            }
        }

        private void Sensor_IsAvailableChanged(object sender, IsAvailableChangedEventArgs e)
        {
            // 如果失败，设置状态文本
            StatusText = _kinectSensor.IsAvailable ? Properties.Resources.RunningStatusText : Properties.Resources.SensorNotAvailableStatusText;
        }

        private void Settings_Click(object sender, RoutedEventArgs e)
        {
            var dialog = new SettingsWindow();
            dialog.ShowDialog();
            if (dialog.DialogResult.Value)
            {
                VideoPlayer.Source = new Uri(Properties.Settings.Default.Video);
                VideoPlayer.Play();
                soundPlayer.SoundLocation = Properties.Settings.Default.Audio;
            }
        }

        private void FullScreen_Click(object sender, RoutedEventArgs e)
        {
            ToolBar.Visibility = Visibility.Hidden;
            VideoPlayer.Margin = new Thickness { Left = 0, Top = 0, Right = 0, Bottom = 0 };
            ViewBox.Height = 300;
            ViewBox.Width = 300;
            Topmost = false;
            WindowStyle = WindowStyle.None;
            WindowState = WindowState.Maximized;
        }

        private void Help_Click(object sender, RoutedEventArgs e)
        {
            new HelpWindow().ShowDialog();
        }

        private void About_Click(object sender, RoutedEventArgs e)
        {
            new AboutWindow().ShowDialog();
        }

        private void Music_Checked(object sender, RoutedEventArgs e)
        {
            if (string.IsNullOrEmpty(soundPlayer.SoundLocation))
            {
                MessageBox.Show("请先配置音频！", "提示", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            else
            {
                // 同步加载声音
                soundPlayer.Load();
                // 循环播放音频
                soundPlayer.PlayLooping();
            }
        }

        private void Music_Unchecked(object sender, RoutedEventArgs e)
        {
            soundPlayer?.Stop();
        }

        private static void Media_Loaded(object sender, RoutedEventArgs e)
        {
            (sender as MediaElement)?.Pause();
        }

        private static void Media_MediaEnded(object sender, RoutedEventArgs e)
        {
            // MediaElement 需要先停止播放才能再开始播放，否则会停在最后一帧不动
            (sender as MediaElement)?.Stop();
            (sender as MediaElement)?.Play();
        }

        private static void Media_Unloaded(object sender, RoutedEventArgs e)
        {
            (sender as MediaElement)?.Stop();
        }
    }
}
